<?php
/*
 * guoba
 * @author Shinsuke Takeuchi <ctrans@gmail.com>
 * @license MIT License
 * @link http://www.karak.jp/
 * @version 1.06
 */

require_once 'imdr.php';
define("MIN_FLOAT", -3.14e+100);

class guoba {
	private $freqtotal;
	private $initdata;
	private $initfn;
	private $self_dir;
	private $trie;
	private $triefn;
	private $start_p;
	private $trans_p;
	private $emit_p;

	#初期化
	public function __construct() {
		$this->self_dir = dirname(__FILE__);
		$this->dat_dir = $this->self_dir . "/dat";
		$this->triefn = $this->dat_dir . "/trie.dat";
		$this->startfn = $this->dat_dir . "/start.dat";
		$this->transfn = $this->dat_dir . "/trans.dat";
		$this->emitfn = $this->dat_dir . "/emit.dat";
		$this->initfn = $this->dat_dir . "/init.dat";
		
		$this->trie = new imdr();
		$this->trie->load($this->triefn);
		$this->start_p = $this->serial_load($this->startfn);
		$this->trans_p = $this->serial_load($this->transfn);
		$this->emit_p = $this->serial_load($this->emitfn);
		
		$this->initdata = $this->serial_load($this->initfn);
		
		$this->freqtotal = $this->initdata["word_freq_total"];
	}

	#単語分割
	public function cut($str, $nohmm=FALSE) {
		$wordlist = $this->get_wordlist($str);
		$route = $this->calc($str, $wordlist);
		$out = array();
		if ($nohmm) {
			$out = $this->NOHMM($str, $route);
			return $out;
		}
		$x = 0;
		$buf = "";
		$enc = "UTF-8";
		$len = mb_strlen($str, $enc);
		while ($x < $len) {
			list($y, $rate) = $route[$x];
			$tail = $y - $x + 1;
			$frag = mb_substr($str, $x, $tail, $enc);
			#1文字に分割された文字列から未登録語を識別
			if ($x == $y) {
				$buf = $buf . $frag;
			} else {
				if ($buf != "") {
					if ($this->trie->getstate($buf) == -1) {
						$split_buf = $this->han_split($buf);
						foreach ($split_buf as $sb) {
							if ($sb == "") {continue;}
							if (preg_match("/^[^一-龠]+$/u", $sb)) {
								$out[] = $sb;
								continue;
							}
							$recognized = $this->HMM($sb);
							foreach ($recognized as $elm) {
								$out[] = $elm;
							}
						}
					} else {
						$out[] = $buf;
					}
					$buf = "";
				}
				$out[] = $frag;
			}
			$x = $y + 1;
		}
		#残ったbufの処理
		if ($buf != "") {
			if($this->trie->getstate($buf) == -1) {
				$split_buf = $this->han_split($buf);
				foreach ($split_buf as $sb) {
					if ($sb == "") {continue;}
					if (preg_match("/^[^一-龠]+$/u", $sb)) {
						$out[] = $sb;
						continue;
					}
					$recognized = $this->HMM($sb);
					foreach ($recognized as $elm) {
						$out[] = $elm;
					}
				}
			} else {
				$out[] = $buf;
			}
		}
		return $out;
	}

	#漢字と漢字以外の文字で文字列を分割する
	public function han_split($str) {
		$bom = "\u{feff}";
		$str = preg_replace("/([A-z0-9%@.\-\/]+)/u", "$bom$1$bom", $str);
		return explode($bom, $str);
	}

	#HMMによる単語分割（この関数だけでも分割は可能（辞書なし））
	public function HMM($str) {
		$enc = "UTF-8";
		$len = mb_strlen($str, $enc);
		$words = array();
		#$strが1文字の場合は未登録語の識別処理は不要
		#$strが英数字の場合も同様
		if ($len == 1 || preg_match("/[A-z0-9%@.\-\/]+/u", $str)) {
			$words[] = $str;
			return $words;
		}
		$path = array();
		$path = $this->viterbi($str);
		$begin = 0;
		$next = 0;
		for ($i = 0; $i<$len; $i++) {
			$char = mb_substr($str, $i, 1, $enc);
			$pos = $path[$i];
			if ($pos == "B") {
				$begin = $i;
			} elseif ($pos == "E") {
				#$count = $i - $begin;
				$words[] = mb_substr($str, $begin, (($i+1)-$begin), $enc);
				$next = $i + 1;
			} elseif ($pos == "S") {
				$words[] = $char;
				$next = $i + 1;
			}
		}
		if ($next < $len) {
			$words[] = mb_substr($str, $next, null, $enc);
		}
		return $words;
	}

	#ビタビ法で未登録語を認識
	public function viterbi($obs) {
		$states = array('B', 'M', 'E', 'S');
		$prev_states = array('B'=>array('E', 'S'), 'M'=>array('M', 'B'), 'E'=>array('B', 'M'), 'S'=>array('E', 'S'));
		$V = array();
		$V[0] = array();
		$path = array();
		$enc = "UTF-8";
		#処理対象の文字列obsの1文字目を使って初期化
		foreach ($states as $key => $state) {
			$y = $state;
			$head = mb_substr($obs, 0, 1, $enc);
			$emit = 0.0;
			if (isset($this->emit_p[$y][$head])) {
				$emit = log($this->emit_p[$y][$head]/$this->emit_p['SUM']);
			} else {
				$emit = MIN_FLOAT;
			}
			$V[0][$y] = $this->start_p[$y] + $emit;
			$path[$y] = $y;
		}
		for ($t = 1; $t < mb_strlen($obs, $enc); $t++) {
			$V[$t] = array();
			$newpath = array();
			$char = mb_substr($obs, $t, 1, $enc);
			foreach ($states as $key => $state) {
				$y = $state;
				if (isset($this->emit_p[$y][$char])) {
					$emit = log($this->emit_p[$y][$char]/$this->emit_p['SUM']);
				} else {
					$emit = MIN_FLOAT;
				}
				$prob = MIN_FLOAT;
				$state = "S";
				#1文字ずつ走査しているので1つ前のインデックスはt-1で得られる
				$previdx = $t - 1;
				foreach ($prev_states[$y] as $y0) {
					$prev = $V[$previdx][$y0];
					$trans = $trans_p[$y0][$y];
					$res = $prev + $trans + $emit;
					if ($res > $prob) {
						$prob = $res;
						$state = $y0;
					}
				}
				#現在のyと1つ前のstate(y0)で一番もっともらしい組み合わせを保存
				$V[$t][$y] = $prob;
				if (is_array($path[$state])) {
					$newpath[$y] = array();
					foreach ($path[$state] as $key => $path_value) {
						$newpath[$y][] = $path_value;
					}
					$newpath[$y][] = $y;
				} else {
					$newpath[$y] = array($path[$state], $y);
				}
			}
			$path = $newpath;
		}
		#最後の1文字はE（終端）かS（独立）なので両者の値を比較
		$last = mb_strlen($obs, $enc) - 1;
		$tail = array('E','S');
		$max = MIN_FLOAT;
		foreach ($tail as $key => $y) {
			$prob = $V[$last][$y];
			if ($prob > $max) {
				$state = $y;
				$max = $prob;
			}
		}
		return $path[$state];
	}

	#末尾から先頭に向けて隣接する2つの単語の出現頻度の合計を求め、
	#合計値が高い組み合わせを候補として保存する
	public function calc($str, $wordlist) {
		$enc = "UTF-8";
		$len = mb_strlen($str, $enc);
		$route = array();
		$route[$len] = array($len, 0.0);
		#後ろから前に向けて走査
		for($idx = ($len - 1); $idx > -1; $idx--) {
			$max = MIN_FLOAT;
			foreach ($wordlist[$idx] as $frag) {
				$surface = $frag['surface'];
				$taginfo = $frag['taginfo'];
				$tailpos = $frag['tailpos'];
				#品詞情報はjson形式（{"v":"1","SUM":"1"}）で保存されているのでデコードする
				$tags = json_decode($taginfo, true);
				#この単語の出現率を求める（SUMはコーパス内の合計出現回数）
				$cur_freq = log($tags["SUM"] / $this->freqtotal);
				$next = $tailpos + 1;
				if (isset($route[$next])) {
					list($tailidx, $prev_freq) = $route[$next];
				} else {
					$prev_freq = 0.0;
				}
				$result = $cur_freq + $prev_freq;
				if ($result >= $max) {
					$max = $result;
					#経路routeには$idxから始まり、$tailposで終わる単語の出現率（$max）を保存する
					$route[$idx] = array($tailpos, $max);
				}
			}
		}
		return $route;
	}

	#未登録語の識別処理なしで出力
	public function NOHMM($str, $route) {
		$x = 0;
		$buf = "";
		$out = array();
		$enc = "UTF-8";
		$len = mb_strlen($str, $enc);
		while ($x < $len) {
			list($y, $rate) = $route[$x];
			$tail = $y - $x + 1;
			$out[] = mb_substr($str, $x, $tail, $enc);
			$x = $y + 1;
		}
		return $out;
	}

	#文字列strのi番目の文字から共通接頭辞を求め、配列に保存
	#DAG（有向非循環グラフ）としても利用するため単語末尾の位置も保存
	public function get_wordlist($str) {
		$enc = "UTF-8";
		$len = mb_strlen($str, $enc);
		$wordlist = array();
		for ($i = 0; $i < $len; $i++) {
			$cur = mb_substr($str, $i, null, $enc);
			$matches = $this->trie->getprefix($cur);
			#$iの位置にある文字が辞書に未登録であった場合
			if (count($matches) == 0) {
				$frag = mb_substr($str, $i, 1, $enc);
				#trieに新規単語として登録する
				$this->trie->add($frag, '{"n":"1","SUM":"1"}');
				$state = $this->trie->getstate($frag);
				$matches[$frag] = $state;
				#$this->freqtotalも更新
				$this->freqtotal += 1;
				$this->initdata["word_freq_total"] = $this->freqtotal;
				$this->save();
			}
			#wordlist構築
			$wordlist[$i] = array();
			$tmplist = array();
			foreach ($matches as $surface => $state) {
				$taginfo = $this->trie->getvalue($state);
				$tailpos = $i + mb_strlen($surface, $enc) - 1;
				#現在の開始位置（i）、単語、品詞情報（各品詞の出現回数）、末尾位置を保存
				$wordlist[$i][] = array('surface' =>$surface, 'taginfo'=>$taginfo, 'tailpos'=>$tailpos);
			}
		}
		return $wordlist;
	}

	#trieに新規単語を登録し（引数：単語、品詞、出現回数 -> 例：tango, n, 10）、
	#BEMSの遷移データも生成する
	#英数字のみで構成される単語は登録しないが、$forceがTRUEの場合は登録する
	public function add($word, $pos, $freq, $force=FALSE) {
		if ($force == FALSE && preg_match("/^[A-z0-9%@.\-\/]+$/u", $word)) {return FALSE;}
		$pos = mb_strtolower($pos);
		if (!preg_match("/^[a-z]+$/u", $pos)) {$pos = "n";}
		$state = $this->trie->getstate($word);
		$arry = array();
		if ($state != -1) {
			#品詞情報はjson形式で保存されているのでデコード
			$json_value = $this->trie->getvalue($state);
			$arry = json_decode($json_value, true);
			if (array_key_exists($pos, $arry)) {
				$arry[$pos] += $freq;
			} else {
				$arry[$pos] = $freq;
			}
			if (array_key_exists('SUM', $arry)) {
				$arry['SUM'] += $freq;
			} else {
				$arry['SUM'] = $arry[$pos];
			}
		} else {
			$arry[$pos] = $freq;
			$arry['SUM'] = $freq;
		}
		$encoded = json_encode($arry);
		$this->trie->add($word, $encoded);
		#遷移データ生成
		$this->emit_p['SUM'] += $this->regist_bems($word);
		#頻度データ更新
		$this->freqtotal += $freq;
		$this->initdata["word_freq_total"] = $this->freqtotal;
		return TRUE;
	}

	#指定された単語が登録されているか否か
	public function exists($word) {
		if ($this->trie->getstate($word) != -1) {
			return TRUE;
		}
		return FALSE;
	}

	#ファイルを読み込んでまとめて単語登録
	#ファイルの書式は、1行に「単語 品詞 頻度」がスペースで区切られているかたちとする
	public function add_from_file($fn) {
		if (!file_exists($fn)) {return FALSE;}
		$fp = fopen($fn,"r");
		while (($line = fgets($fp))) {
			#trim
			$line = preg_replace('/\A[\x00\s]++|[\x00\s]++\z/u', '', $line);
			if($line == "") {continue;}
			list($word, $pos, $freq) = explode(" ", $line);
			#品詞が複数登録されている場合（例：n-v）
			$wclass = explode("-", $pos);
			foreach ($wclass as $wc) {
				$this->add($word, $wc, $freq);
			}
		}
		fclose($fp);
		$this->save();
		return TRUE;
	}

	#trieの単語データを削除する
	public function remove ($word) {
		$state = $this->trie->getstate($word);
		if ($state != -1) {
			#品詞情報はjson形式で保存されているのでデコード
			$json_value = $this->trie->getvalue($state);
			$arry = array();
			$arry = json_decode($json_value, true);
			if (array_key_exists('SUM', $arry)) {
				$freq = $arry['SUM'];
				$this->freqtotal -= $freq;
				$this->initdata["word_freq_total"] = $this->freqtotal;
			}
		}
		return $this->trie->remove($word); 
	}

	#単語分割済みのコーパス（教師データ）を読み込み、
	#辞書データとemitデータ（BEMSの遷移）を同時に学習する
	#教師データの書式は「単語/品詞 単語/品詞」のように
	#人民日報コーパスとほぼ同様の書式（ただし、単語間の区切りのスペースは1つだけ）
	public function learn($fn) {
		set_time_limit(600);
		#ゼロから学習させ直したい場合は以下4行をコメントアウト
		#unset($this->emit_p);
		#unset($this->trie);
		#$this->emit_p = array("B"=>array(), "E"=>array(), "M"=>array(), "S"=>array(), "SUM"=>0);
		#$this->trie = new imdr();
		if (!file_exists($fn)) {return FALSE;}
		if (!isset($this->emit_p['SUM'])) {$this->emit_p['SUM'] = 0;}
		$fp = fopen($fn,"r");
		while ($line = fgetss($fp)) {
			#単語前後の余分な空白を除去
			$line = preg_replace('/\A[\x00\s]++|[\x00\s]++\z/u', '', $line);
			if($line == "") {continue;}
			$words = explode(" ", $line);
			#これまでに処理した全文字数を取得
			$char_total = $this->emit_p['SUM'];
			foreach($words as $word) {
				list($w, $pos) = explode("/", $word);
				#辞書データ登録
				$this->add($w, $pos, 1);
			}
		}
		fclose($fp);
		return TRUE;
	}

	#遷移データ生成
	public function regist_bems($word) {
		$enc = "UTF-8";
		$len = mb_strlen($word, $enc);
		#1文字の単語
		if ($len == 1) {
			$pname = "S";
			#すでにキー$wordが存在している場合
			if (isset($this->emit_p[$pname][$word])) {
				$emit = $this->emit_p[$pname][$word];
				$emit++;
			} else {
				$emit = 1;
			}
			$this->emit_p[$pname][$word] = $emit;
			return $len;
		}
		for ($i = 0; $i < $len; $i++) {
			$char = mb_substr($word, $i, 1, $enc);
			if ($i == 0) {
				$pname = "B";
			} elseif ($i == ($len - 1)) {
				$pname = "E";
			} else {
				$pname = "M";
			}
			if (isset($this->emit_p[$pname][$char])) {
				$emit = $this->emit_p[$pname][$char];
				$emit++;
			} else {
				$emit = 1;
			}
			$this->emit_p[$pname][$char] = $emit;
		}
		return $len;
	}

	#指定されたファイルを読み込み
	public function load($fn) {
		if (!file_exists($fn)) {return;}
		$content = file_get_contents($fn);
		return $content;
	}

	#trieとemitを保存する
	public function save() {
		$this->trie->save($this->triefn);
		$this->serial_save($this->emitfn, $this->emit_p);
		$this->serial_save($this->initfn, $this->initdata);
		return;
	}

	#データ（配列など）をシリアライズして保存
	public function serial_save($fn, $dat) {
		$s = serialize($dat);
		$fp = fopen($fn, "w");
		fwrite($fp, $s);
		fclose($fp);
	}

	#ファイルからシリアライズされたデータを読み込み
	public function serial_load($fn) {
		if (!file_exists($fn)) {return;}
		$content = file_get_contents($fn);
		$dummy = unserialize($content);
		return $dummy;
	}

}


?>